@@ -34,8 +34,8 @@ func (c *mockedConfig) Options() config.OptionsConfiguration {
 	return config.OptionsConfiguration{}
 }
 
-func (c *mockedConfig) Replace(cfg config.Configuration) error {
-	return nil
+func (c *mockedConfig) Replace(cfg config.Configuration) (config.Waiter, error) {
+	return nil, nil
 }
 
 func (c *mockedConfig) Subscribe(cm config.Committer) {}
@@ -48,12 +48,12 @@ func (c *mockedConfig) Devices() map[protocol.DeviceID]config.DeviceConfiguratio
 	return nil
 }
 
-func (c *mockedConfig) SetDevice(config.DeviceConfiguration) error {
-	return nil
+func (c *mockedConfig) SetDevice(config.DeviceConfiguration) (config.Waiter, error) {
+	return nil, nil
 }
 
-func (c *mockedConfig) SetDevices([]config.DeviceConfiguration) error {
-	return nil
+func (c *mockedConfig) SetDevices([]config.DeviceConfiguration) (config.Waiter, error) {
+	return nil, nil
 }
 
 func (c *mockedConfig) Save() error {
