@@ -8,12 +8,4 @@
 
 package osutil
 
-func HideFile(path string) error {
-	return nil
-}
-
-func ShowFile(path string) error {
-	return nil
-}
-
 func HideConsole() {}
