@@ -16,7 +16,7 @@ import (
 )
 
 const (
-	// BlockSize is the standard ata block size (128 KiB)
+	// BlockSize is the standard data block size (128 KiB)
 	BlockSize = 128 << 10
 
 	// MaxMessageLen is the largest message size allowed on the wire. (500 MB)
@@ -127,7 +127,7 @@ type message interface {
 
 type asyncMessage struct {
 	msg  message
-	done chan struct{} // done closes when we're done marshalling the message and it's contents can be reused
+	done chan struct{} // done closes when we're done marshalling the message and its contents can be reused
 }
 
 const (
